import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mobx/mobx.dart';

import 'package:endlich/src/scheduler.dart';

part 'machine.freezed.dart';

/// A callback to be executed when a state transitions to another state.
/// This should be merely used to clean up some local state created by [StateMachine.enterState].
typedef OffState = void Function();

/// Send an event to the [StateMachine]
typedef SendEvent = void Function(StateMachineEvent event);

/// A function to be called when an event was sent to the [StateMachine].
/// [willTransition] indicates wheter the StateMachine will be in a new status/state
/// after the event.
typedef SessionEventListener = void Function(
    StateMachineEvent event, bool willTransition);

/// Starts the StateMachine in the state returned by [getInitialState] by default.
///
/// The machine can be started in a different state by using [startIn].
/// In this case [getInitialState] will never be called.
///
/// In any case, it will call [enterState].
///
/// All listeners will be notified
typedef StateMachineStart<State> = void Function(
    {State Function(SendEvent send)? startIn});

/// Immediately stops the StateMachine even if there are events to be executed.
/// It will also dispose the current session and so will execute all [StateMachineSession.onLeave] callbacks.
///
/// All listeners will be notified
typedef StateMachineStop = void Function();

/// The current Status of the StateMachine
@freezed
class StateMachineStatus<State> with _$StateMachineStatus<State> {
  /// Default status of a [StateMachine]. Indicates that the machine has never
  /// been started or stopped before.
  ///
  /// Use [start] to start the machine.
  const factory StateMachineStatus.idle(StateMachineStart<State> start) =
      StateMachineStatusIdle<State>;

  /// The [StateMachine] is stopped.There is no active [StateMachineSession].
  ///
  /// Use [start] to start the machine.
  const factory StateMachineStatus.stopped(
          State lastState, StateMachineStart<State> start) =
      StateMachineStatusStopped<State>;

  /// The [StateMachine] is running and has a valid [State]
  ///
  /// Use [stop] to stop the machine.
  const factory StateMachineStatus.running(State state, StateMachineStop stop) =
      StateMachineStatusRunning<State>;
}

/// Send and event to the StateMachine
/// for example:
/// StateMachineEvent('setName', 'Luke');
class StateMachineEvent {
  StateMachineEvent(this.name, this.data);

  /// The name of the event
  final String name;

  /// The Payload
  final Object? data;

  @override
  String toString() => 'StateMachineEvent(name: $name, data: $data)';

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;

    return other is StateMachineEvent &&
        other.name == name &&
        other.data == data;
  }

  @override
  int get hashCode => name.hashCode ^ data.hashCode;
}

/// Base class that handles state transitions.
abstract class StateMachine<State> {
  StateMachine({bool autostart = false}) {
    if (autostart) {
      _start();
    }
  }

  static final ReactiveContext _ctx =
      ReactiveContext(config: ReactiveConfig.main);

  final Scheduler _scheduler = Scheduler();

  bool _disposed = false;

  late final ActionController _ctrl =
      ActionController(context: _ctx, name: '$runtimeType.ActionController');

  late final Observable<StateMachineStatus<State>> _status$ =
      Observable<StateMachineStatus<State>>(
    StateMachineStatus<State>.idle(_start),
    context: _ctx,
    equals: (p0, p1) => false,
    name: '$runtimeType.status',
  );

  final List<ReactionDisposer> _watchDisposer = [];

  /// Whether [dispose] was called or not.
  bool get isDisposed => _disposed;

  /// Current active session
  @protected
  @visibleForTesting
  late StateMachineSession session;

  /// The current active [StateMachineStatus].
  /// Usually you don't want to access this property and rather add a callback in
  /// [StateMachine.watch] to listen for status changes.
  ///
  /// Technically this is a mobx Observable so it supports all mobx functions.
  StateMachineStatus<State> get status {
    assert(_debugIsDisposed(), '');
    return _status$.value;
  }

  bool get hasListener {
    assert(_debugIsDisposed(), '');
    return _watchDisposer.isNotEmpty;
  }

  bool _debugIsDisposed() {
    assert(() {
      if (_disposed) {
        throw StateError('''
Tried to use $runtimeType after `dispose` was called.
Consider checking `isDisposed`.
''');
      }
      return true;
    }(), '');
    return true;
  }

  @mustCallSuper
  void dispose() {
    _stop();

    _watchDisposer
      ..forEach((cb) => cb())
      ..clear();

    _disposed = true;
  }

  /// Return the initial State of the [StateMachine]
  @protected
  State getInitialState(SendEvent send);

  /// Execute side effects when entering a new state.
  /// It will also execute on the initial first state.
  ///
  /// [state] is the current active state in the [StateMachine].
  /// [session] is the current active [StateMachineSession] in the [StateMachine].
  @protected
  void enterState(State state, StateMachineSession session);

  /// Taken the current active [state] and the [event] send to the [StateMachine]
  /// will reult a new state.
  ///
  /// Return [state] if there is no transition to take because the current state does not support
  /// the event. This will prevent a transition and so will not call any listener or [enterState]
  @protected
  State transition(State state, StateMachineEvent event, SendEvent send);

  /// Create a custom [StateMachineSession].
  /// This could be useful if you want to add custom logic to the session.
  @protected
  StateMachineSession createSession(State state, SendEvent send) {
    return StateMachineSession(send);
  }

  void _start({State Function(SendEvent send)? startIn}) {
    assert(_debugIsDisposed(), '');
    void _start() {
      final send = _createSend();
      final initialState = startIn?.call(send) ?? getInitialState(send);
      session = createSession(initialState, send);
      final info = _ctrl.startAction(name: '$runtimeType.start');
      try {
        _scheduler.schedule(() {
          if (!session.isActive) return;
          _status$.value =
              StateMachineStatus<State>.running(initialState, _stop);

          enterState(initialState, session);
        });
      } finally {
        _ctrl.endAction(info);
      }
    }

    _status$.value
        .map(idle: (_) => _start(), stopped: (_) => _start(), running: (_) {});
  }

  void _stop() {
    assert(_debugIsDisposed(), '');
    _status$.value.map(
        idle: (_) {},
        stopped: (_) {},
        running: (running) {
          final info = _ctrl.startAction(name: '$runtimeType.stop');
          try {
            _status$.value =
                StateMachineStatus<State>.stopped(running.state, _start);
            session.dispose();
          } finally {
            _ctrl.endAction(info);
          }
        });
  }

  SendEvent _createSend() {
    var didTransition = false;
    return (StateMachineEvent event) {
      assert(_debugIsDisposed(), '');
      _status$.value.map(
        idle: (_) {},
        stopped: (_) {},
        running: (running) {
          if (_disposed || didTransition) return;

          final currentSession = session;

          final send = _createSend();
          final nextState = transition(running.state, event, send);
          final willTransition = nextState != running.state;
          currentSession.executeListener(event, willTransition);

          // this can happen if an event listener did send a new event
          if (!currentSession.isActive || !willTransition) return;

          final nextSession = createSession(nextState, send);
          didTransition = true;
          currentSession.dispose();
          session = nextSession;

          final info = _ctrl.startAction(name: '$runtimeType.send');
          try {
            _scheduler.schedule(() {
              if (!nextSession.isActive ||
                  _status$.value is! StateMachineStatusRunning<State>) return;
              _status$.value =
                  StateMachineStatus<State>.running(nextState, _stop);

              enterState(nextState, nextSession);
            });
          } finally {
            _ctrl.endAction(info);
          }
        },
      );
    };
  }

  /// Watch the StateMachine status. The [cb] will immediately receive the current
  /// status of the machine.
  /// This can be prevented by setting [fireImmediately] to false (Default: true)
  @mustCallSuper
  void Function() watch(
    void Function(StateMachineStatus<State> status) cb, {
    bool fireImmediately = true,
  }) {
    assert(_debugIsDisposed(), '');
    final disposer = reaction<StateMachineStatus<State>>(
      (_) => _status$.value,
      cb,
      context: _ctx,
      equals: (_, __) => false,
      fireImmediately: fireImmediately,
      name: '$runtimeType.watch',
    );

    _watchDisposer.add(disposer);

    return () {
      if (disposer.reaction.isDisposed) return;
      _watchDisposer.removeWhere((d) => d == disposer);
      disposer();
    };
  }
}

/// Reflects the current state in a [StateMachine].
class StateMachineSession {
  StateMachineSession(this._send);

  final void Function(StateMachineEvent event) _send;
  final List<OffState> _onLeave = [];
  final List<SessionEventListener> _eventListener = [];

  bool _disposed = false;

  /// Inidicates if the current session is still active in the StateMachine.
  /// This is especially useful if you are going to execute an async task in
  /// [StateMachine.enterState].
  ///
  /// Do not use the [StateMachineSession] or send events to the [StateMachine]
  /// after [isActive] is false.
  bool get isActive => !_disposed;

  bool _debugIsDisposed() {
    assert(() {
      if (_disposed) {
        throw StateError('''
Tried to use $runtimeType after `dispose` was called.
Consider checking `isActive`.
''');
      }
      return true;
    }(), '');
    return true;
  }

  @protected
  @visibleForTesting
  void executeListener(StateMachineEvent event, bool willTransition) {
    assert(_debugIsDisposed());
    final listener = [..._eventListener];
    for (var cb in listener) {
      cb(event, willTransition);
    }
  }

  @mustCallSuper
  void dispose() {
    _disposed = true;

    _eventListener.clear();
    final leave = [..._onLeave];
    _onLeave.clear();
    for (var cb in leave) {
      cb();
    }
  }

  /// Send any event to the StateMachine.
  void send(StateMachineEvent event) {
    assert(_debugIsDisposed());
    if (_disposed) return;
    _send(event);
  }

  /// Execute a [cb] when the current session is disposed.
  /// The primary usage is to clean up local state that was created in [StateMachine.enterState].
  /// For example to stop/cancel a http call.
  ///
  /// The [cb] will be executed if the StateMachine left its current status/state.
  /// Possible reasons are: Transition to another state, Stop or dispose of the StateMachine
  ///
  /// *IMPORTANT*: Do not use the current session or send events in this callback.
  void onLeave(void Function() cb) {
    assert(_debugIsDisposed());
    if (_disposed) return;
    _onLeave.add(cb);
  }

  /// Allows to listen to all events send to the [StateMachine].
  /// Use this if you want to add a side effect to an event.
  ///
  /// The [cb] will be called _BEFORE_ the transition occurs.
  ///
  /// The [cb] will automatically be removed when the session is disposed.
  void listenToEvent(SessionEventListener cb) {
    assert(_debugIsDisposed());
    if (_disposed) return;
    _eventListener.add(cb);
  }
}
