import 'package:endlich/endlich.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';
import 'machine_test.mocks.dart';
import 'models.dart';

class InitialState extends Mock {
  TestState call(SendEvent send);
}

class Transition extends Mock {
  TestState call(TestState state, StateMachineEvent event, SendEvent send);
}

class EnterState extends Mock {
  void call(TestState state, StateMachineSession session);
}

class Watch extends Mock {
  void call(StateMachineStatus<TestState> status);
}

class OnLeave extends Mock {
  void call();
}

@GenerateMocks([OnLeave, Watch, InitialState, Transition, EnterState])
void main() {
  TypeMatcher<T> isStateOf<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }

    return matcher;
  }

  StateMachineStatusRunning<TestState> assumeIsRunning<T extends TestState>(
      TestMachine m) {
    expect(
        m.status,
        isA<StateMachineStatusRunning<TestState>>()
            .having((s) => s.state, 'running.state', isStateOf<T>()));
    return m.status as StateMachineStatusRunning<TestState>;
  }

  StateMachineStatusIdle<TestState> assumeIsIdle(TestMachine m) {
    expect(m.status, isA<StateMachineStatusIdle<TestState>>());
    return m.status as StateMachineStatusIdle<TestState>;
  }

  StateMachineStatusStopped<TestState> assumeIsStopped<T extends TestState>(
      TestMachine m) {
    expect(
        m.status,
        isA<StateMachineStatusStopped<TestState>>()
            .having((s) => s.lastState, 'stopped.state', isStateOf<T>()));
    return m.status as StateMachineStatusStopped<TestState>;
  }

  TypeMatcher<StateMachineStatusRunning<TestState>>
      isStatusRunning<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }

    return isA<StateMachineStatusRunning<TestState>>()
        .having((s) => s.state, 'running.state', matcher);
  }

  TypeMatcher<StateMachineStatusStopped<TestState>>
      isStatusStopped<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }
    return isA<StateMachineStatusStopped<TestState>>()
        .having((s) => s.lastState, 'running.state', matcher);
  }

  TypeMatcher<StateMachineStatusIdle<TestState>> isStatusIdle() {
    return isA<StateMachineStatusIdle<TestState>>();
  }

  test('correct order', () {
    final foo = TestState.foo(0, (incr) {});
    var transitions = [
      foo,
      TestState.bar(1, (incr) {}),
      TestState.baz(2, (incr) {}),
    ];

    final MockTransition mockTransition = MockTransition();
    final MockInitialState mockInitialState = MockInitialState();
    final MockEnterState mockEnterState = MockEnterState();
    final fooLeave = MockOnLeave();

    when(mockInitialState(any)).thenReturn(foo);
    when(mockTransition(any, any, any))
        .thenAnswer((_) => transitions.removeAt(0));

    final m = TestMachine(
      configuredEnterState: (state, session) {
        mockEnterState(state, session);
        if (state is Foo) {
          session.onLeave(fooLeave);
        }
      },
      configuredInitialState: (send) {
        return mockInitialState(send);
      },
      configuredTransition: (state, event, send) {
        return mockTransition(state, event, send);
      },
    );
    final watcher = MockWatch();
    m.watch(watcher);

    assumeIsIdle(m).start();
    m.send(StateMachineEvent('guarded', null));
    m.send(StateMachineEvent('notGuarded', null));
    assumeIsRunning(m).stop();
    assumeIsStopped(m).start();
    m.dispose();

    verifyInOrder([
      watcher(argThat(isStatusIdle())),
      mockInitialState(any),
      mockEnterState(argThat(isStateOf<Foo>()), any),
      watcher(argThat(isStatusRunning<Foo>())),

      /// guarded transition on 'guarded' event => return foo
      mockTransition(
          argThat(isStateOf<Foo>()),
          argThat(isA<StateMachineEvent>()
              .having((p0) => p0.name, 'event name', 'guarded')),
          any),

      mockTransition(
          argThat(isStateOf<Foo>()),
          argThat(isA<StateMachineEvent>()
              .having((p0) => p0.name, 'event name', 'notGuarded')),
          any),
      fooLeave(),
      mockEnterState(argThat(isStateOf<Bar>()), any),
      watcher(argThat(isStatusRunning<Bar>())),
      watcher(argThat(isStatusStopped<Bar>())),
      mockInitialState(any),
      mockEnterState(argThat(isStateOf<Foo>()), any),
      watcher(argThat(isStatusRunning<Foo>())),
      fooLeave(),
      watcher(argThat(isStatusStopped<Foo>())),
    ]);

    verifyNoMoreInteractions(mockInitialState);
    verifyNoMoreInteractions(mockTransition);
    verifyNoMoreInteractions(mockEnterState);
    verifyNoMoreInteractions(watcher);
    verifyNoMoreInteractions(fooLeave);
  });

  test('leaving state in enterState', () {
    var transitions = [
      TestState.bar(1, (incr) {}),
      TestState.baz(2, (incr) {}),
    ];

    final MockTransition mockTransition = MockTransition();
    final MockInitialState mockInitialState = MockInitialState();
    final MockEnterState mockEnterState = MockEnterState();
    final watcher = MockWatch();
    final fooLeave = MockOnLeave();
    final barLeave = MockOnLeave();

    when(mockInitialState(any)).thenReturn(TestState.foo(0, (incr) {}));
    when(mockTransition(any, any, any))
        .thenAnswer((_) => transitions.removeAt(0));

    final m = TestMachine(
      configuredEnterState: (state, session) {
        mockEnterState(state, session);
        if (state is Foo) {
          session.onLeave(fooLeave);
          session.send(StateMachineEvent('anything', null));
        } else if (state is Bar) {
          session.onLeave(barLeave);
          session.send(StateMachineEvent('anything', null));
        }
      },
      configuredInitialState: (send) {
        return mockInitialState(send);
      },
      configuredTransition: (state, event, send) {
        return mockTransition(state, event, send);
      },
    );
    m.watch(watcher);
    assumeIsIdle(m).start();

    verifyInOrder([
      watcher(argThat(isStatusIdle())),
      mockInitialState(any),
      mockEnterState(argThat(isStateOf<Foo>()), any),
      mockTransition(
          argThat(isStateOf<Foo>()), argThat(isA<StateMachineEvent>()), any),
      fooLeave(),
      mockEnterState(argThat(isStateOf<Bar>()), any),
      mockTransition(
          argThat(isStateOf<Bar>()), argThat(isA<StateMachineEvent>()), any),
      barLeave(),
      mockEnterState(argThat(isStateOf<Baz>()), any),
      watcher(argThat(isStatusRunning<Baz>()))
    ]);

    verifyNoMoreInteractions(mockInitialState);
    verifyNoMoreInteractions(mockTransition);
    verifyNoMoreInteractions(mockEnterState);
    verifyNoMoreInteractions(watcher);
    verifyNoMoreInteractions(fooLeave);
    verifyNoMoreInteractions(barLeave);
  });
}

class TestMachine extends StateMachine<TestState> {
  TestMachine({
    required this.configuredEnterState,
    required this.configuredTransition,
    required this.configuredInitialState,
  }) : super(autostart: false);

  final TestState Function(
          TestState state, StateMachineEvent event, SendEvent send)
      configuredTransition;

  final void Function(TestState state, StateMachineSession session)
      configuredEnterState;

  final TestState Function(SendEvent send) configuredInitialState;

  @override
  void enterState(TestState state, StateMachineSession session) {
    configuredEnterState(state, session);
  }

  @override
  TestState getInitialState(SendEvent send) {
    return configuredInitialState(send);
  }

  @override
  TestState transition(
      TestState state, StateMachineEvent event, SendEvent send) {
    return configuredTransition(state, event, send);
  }

  void send(StateMachineEvent event) {
    session.send(event);
  }
}
