import 'package:endlich/endlich.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import 'state_machine_session_test.mocks.dart';

class Send extends Mock {
  void call(StateMachineEvent event);
}

class MockOnLeave extends Mock {
  void call();
}

class EventListener extends Mock {
  void call(StateMachineEvent event, bool willTransition);
}

@GenerateMocks([Send, EventListener])
void main() {
  Matcher isDisposedError = throwsA(isStateError.having((p0) => p0.message,
      'disposed error', contains('Consider checking `isActive`')));

  test('isActive', () {
    final send = MockSend();
    final session = StateMachineSession(send);
    expect(session.isActive, isTrue);
    session.dispose();
    expect(session.isActive, isFalse);
  });

  group('onLeave', () {
    test('callbacks will be executed on dispose', () {
      final send = MockSend();
      final session = StateMachineSession(send);
      final m1 = MockOnLeave();
      final m2 = MockOnLeave();
      session
        ..onLeave(m1)
        ..onLeave(m2);

      session.dispose();
      verifyInOrder([m1(), m2()]);
      verifyNoMoreInteractions(m1);
      verifyNoMoreInteractions(m2);
    });

    test('will assert while trying to add new callback to disposed session',
        () {
      final send = MockSend();
      final session = StateMachineSession(send)..dispose();

      expect(() => session.onLeave(() {}), isDisposedError);
    });

    test('the session is no longer active in callback', () {
      final send = MockSend();
      final session = StateMachineSession(send);
      var active = session.isActive;
      session.onLeave(() {
        active = session.isActive;
      });
      session.dispose();

      expect(active, isFalse);
    });
  });

  group('event listener', () {
    test('execute', () {
      final send = MockSend();
      final session = StateMachineSession(send);
      final l1 = MockEventListener();
      final l2 = MockEventListener();
      session
        ..listenToEvent(l1)
        ..listenToEvent(l2)
        ..executeListener(StateMachineEvent('foo', null), false)
        ..executeListener(StateMachineEvent('bar', null), true);
      verifyInOrder([
        l1(
            argThat(isA<StateMachineEvent>()
                .having((p0) => p0.name, 'event name', 'foo')),
            argThat(isFalse)),
        l2(
            argThat(isA<StateMachineEvent>()
                .having((p0) => p0.name, 'event name', 'foo')),
            argThat(isFalse)),
        l1(
            argThat(isA<StateMachineEvent>()
                .having((p0) => p0.name, 'event name', 'bar')),
            argThat(isTrue)),
        l2(
            argThat(isA<StateMachineEvent>()
                .having((p0) => p0.name, 'event name', 'bar')),
            argThat(isTrue)),
      ]);
      verifyNoMoreInteractions(l1);
      verifyNoMoreInteractions(l2);
    });

    test('will assert trying to execute if disposed', () {
      final send = MockSend();
      final session = StateMachineSession(send);
      expect(
          () => session
            ..listenToEvent((event, willTransition) {})
            ..dispose()
            ..executeListener(StateMachineEvent('foo', null), false),
          isDisposedError);
    });

    test('will assert while trying to add new callback to disposed session',
        () {
      final send = MockSend();
      final session = StateMachineSession(send)..dispose();

      expect(() => session.listenToEvent((event, willTransition) {}),
          isDisposedError);
    });
  });

  group('send', () {
    test('event', () {
      final send = MockSend();
      final session = StateMachineSession(send);

      session.send(StateMachineEvent('foo', null));
      verify(send(argThat(isA<StateMachineEvent>()
          .having((p0) => p0.name, 'event name', 'foo'))));
    });

    test('will assert if already disposed', () {
      final send = MockSend();
      final session = StateMachineSession(send)..dispose();
      expect(
          () => session.send(StateMachineEvent('foo', null)), isDisposedError);
    });
  });
}
