import 'package:endlich/src/machine.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import 'models.dart';
import 'state_machine_test.mocks.dart';

class EnterState extends Mock {
  void Function()? call(TestState state, StateMachineSession session);
}

class EffectDispose extends Mock {
  void call();
}

class Watch extends Mock {
  void call(StateMachineStatus<TestState> status);
}

class EventListener extends Mock {
  void call(StateMachineEvent event);
}

class EventListenerInSession extends Mock {
  void call(StateMachineEvent event, bool willTransition, TestState state,
      StateMachineSession session);
}

@GenerateMocks(
    [EnterState, EffectDispose, Watch, EventListener, EventListenerInSession])
void main() {
  Matcher isDisposedError = throwsA(isStateError.having((p0) => p0.message,
      'disposed error', contains('Consider checking `isDisposed`')));

  TypeMatcher<T> isStateOf<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }

    return matcher;
  }

  T assumeInState<T extends TestState>(TestMachine m) {
    expect(
        m.status,
        isA<StateMachineStatusRunning<TestState>>()
            .having((s) => s.state, 'running.state', isStateOf<T>()));
    return (m.status as StateMachineStatusRunning<TestState>).state as T;
  }

  StateMachineStatusRunning<TestState> assumeIsRunning<T extends TestState>(
      TestMachine m) {
    expect(
        m.status,
        isA<StateMachineStatusRunning<TestState>>()
            .having((s) => s.state, 'running.state', isStateOf<T>()));
    return m.status as StateMachineStatusRunning<TestState>;
  }

  StateMachineStatusIdle<TestState> assumeIsIdle(TestMachine m) {
    expect(m.status, isA<StateMachineStatusIdle<TestState>>());
    return m.status as StateMachineStatusIdle<TestState>;
  }

  StateMachineStatusStopped<TestState> assumeIsStopped<T extends TestState>(
      TestMachine m) {
    expect(
        m.status,
        isA<StateMachineStatusStopped<TestState>>()
            .having((s) => s.lastState, 'stopped.state', isStateOf<T>()));
    return m.status as StateMachineStatusStopped<TestState>;
  }

  TypeMatcher<StateMachineStatusRunning<TestState>>
      isStatusRunning<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }

    return isA<StateMachineStatusRunning<TestState>>()
        .having((s) => s.state, 'running.state', matcher);
  }

  TypeMatcher<StateMachineStatusStopped<TestState>>
      isStatusStopped<T extends TestState>({
    int? count,
  }) {
    var matcher = isA<T>();

    if (null != count) {
      matcher = matcher.having((s) => s.count, 'state.count', count);
    }
    return isA<StateMachineStatusStopped<TestState>>()
        .having((s) => s.lastState, 'running.state', matcher);
  }

  TypeMatcher<StateMachineStatusIdle<TestState>> isStatusIdle() {
    return isA<StateMachineStatusIdle<TestState>>();
  }

  group('Status', () {
    test('initial depends on autostart', () {
      var m = TestMachine(autostart: false);
      expect(m.status, isStatusIdle());

      m = TestMachine(autostart: true);
      expect(m.status, isStatusRunning());
    });

    test('changes if started/stopped', () {
      var m = TestMachine(autostart: false);
      final watcher = MockWatch();
      m.watch(watcher);
      assumeIsIdle(m).start();
      assumeInState<Foo>(m).next(33);
      assumeIsRunning<Bar>(m).stop();
      assumeIsStopped<Bar>(m).start();
      m.dispose();

      verifyInOrder([
        watcher(argThat(isStatusIdle())),
        watcher(argThat(isStatusRunning<Foo>())),
        watcher(argThat(isStatusRunning<Bar>(count: 33))),
        watcher(argThat(isStatusStopped<Bar>(count: 33))),
        watcher(argThat(isStatusRunning<Foo>())),
        watcher(argThat(isStatusStopped<Foo>(count: 0))),
      ]);
      verifyNoMoreInteractions(watcher);
    });

    test('cannot be started/stopped when machine is disposed', () {
      expect(
          () => assumeIsIdle(TestMachine(autostart: false)..dispose())..start(),
          isDisposedError);
      expect(
          () =>
              assumeIsRunning(TestMachine(autostart: true)..dispose())..stop(),
          isDisposedError);
    });

    test('cannot be accessed if disposed', () {
      expect(
          () => TestMachine(autostart: false)
            ..dispose()
            ..status,
          isDisposedError);
    });

    test('can be stopped at any time, even when events are queued', () {
      var m = TestMachine(
        autostart: false,
        configuredEnterState: (machine, state, session) {
          state.mapOrNull(foo: (foo) {
            foo.next(1);
            assumeIsRunning(machine).stop();
          });
        },
      );
      final watcher = MockWatch();
      m.watch(watcher);
      assumeIsIdle(m).start();
      verifyInOrder([
        watcher(argThat(isStatusIdle())),
        watcher(argThat(isStatusStopped<Foo>(count: 0))),
      ]);
      verifyNoMoreInteractions(watcher);
    });

    test('start in different state', () {
      var m = TestMachine(autostart: false);
      assumeIsIdle(m)
          .start(startIn: (send) => m.createBaz(send: send, count: 100));
      expect(m.status, isStatusRunning<Baz>(count: 100));
    });
  });

  group('transition on event', () {
    test('no longer possible on disposed machine', () {
      final m = TestMachine();
      final foo = assumeInState<Foo>(m);
      m.dispose();
      expect(() => foo.next(33), isDisposedError);
    });

    test('to other state', () {
      final m = TestMachine();

      assumeInState<Foo>(m).next(1);
      expect(m.status, isStatusRunning<Bar>(count: 1));

      assumeInState<Bar>(m).next(1);
      expect(m.status, isStatusRunning<Baz>(count: 2));
    });

    test('is no longer possible for previous state', () {
      final m = TestMachine();
      final foo = assumeInState<Foo>(m);
      foo.next(0);
      foo.next(33);
      expect(m.status, isStatusRunning<Bar>(count: 0));
    });
  });

  group('Session', () {
    group('isActive', () {
      test('will become inactive after transition to other state', () {
        final m = TestMachine();
        final StateMachineSession session = m.debugSession;

        expect(session.isActive, isTrue);
        assumeInState<Foo>(m).next(0);
        expect(session.isActive, isFalse);
      });
    });
  });

  group('enterState', () {
    TypeMatcher<StateMachineSession> isSession({
      bool? isActive,
    }) {
      var matcher = isA<StateMachineSession>();

      if (null != isActive) {
        matcher = matcher.having((s) => s.isActive, 'isActive', isActive);
      }

      return matcher;
    }

    test('will be executed on initial state', () {
      final mockEnterState = MockEnterState();
      when(mockEnterState(any, any)).thenAnswer((realInvocation) => null);
      final m = TestMachine(
          configuredEnterState: (machine, state, session) =>
              mockEnterState(state, session));
      verify(mockEnterState(argThat(isStateOf<Foo>()), argThat(isSession())))
          .called(1);
    });

    test('has an isActive state to check', () {
      late StateMachineSession sessionFoo;
      late StateMachineSession sessionBar;
      late TestState s;
      final m = TestMachine(configuredEnterState: (machine, state, session) {
        s = state;
        if (state is Foo) {
          sessionFoo = session;
        } else if (state is Bar) {
          sessionBar = session;
        }
      });
      expect(sessionFoo.isActive, isTrue);
      s.next(1);
      expect(sessionFoo.isActive, isFalse);
      expect(sessionBar.isActive, isTrue);
      m.dispose();
      expect(sessionBar.isActive, isFalse);
    });

    test(
        'can add a callback function which will be executed when state is going to be left',
        () {
      final fooOff = MockEffectDispose();
      final barOff = MockEffectDispose();

      final m = TestMachine(
        configuredEnterState: (machine, state, session) =>
            state.mapOrNull<void>(
          foo: (value) {
            session.onLeave(fooOff);
          },
          bar: (value) {
            session.onLeave(barOff);
          },
        ),
      );

      assumeInState<Foo>(m).next(0);
      verify(fooOff()).called(1);
      verifyNoMoreInteractions(fooOff);
      verifyZeroInteractions(barOff);

      assumeInState<Bar>(m).next(0);
      verify(barOff()).called(1);
      verifyNoMoreInteractions(barOff);
    });

    test('will be executed in correct order with correct state', () {
      final enterState = MockEnterState();
      final fooEffectOff = MockEffectDispose();
      final barEffectOff = MockEffectDispose();
      final bazEffectOff = MockEffectDispose();
      when(enterState(argThat(isStateOf<Foo>()), any))
          .thenAnswer((realInvocation) => fooEffectOff);
      when(enterState(argThat(isStateOf<Bar>()), any))
          .thenAnswer((realInvocation) => barEffectOff);

      when(enterState(argThat(isStateOf<Baz>()), any))
          .thenAnswer((realInvocation) => bazEffectOff);
      final watcher = MockWatch();

      final m = TestMachine(
        configuredEnterState: (machine, state, session) =>
            state.mapOrNull<OffState?>(
          foo: (value) {
            enterState(state, session);
            session.onLeave(fooEffectOff);
          },
          bar: (value) {
            enterState(state, session);
            session.onLeave(barEffectOff);
          },
          baz: (value) {
            enterState(state, session);
            session.onLeave(bazEffectOff);
          },
        ),
      );

      m.watch(watcher, fireImmediately: true);

      assumeInState<Foo>(m).next(1);
      assumeInState<Bar>(m).next(1);
      assumeInState<Baz>(m).next(1);

      verifyInOrder([
        enterState(argThat(isStateOf<Foo>(count: 0)), argThat(isSession())),
        watcher(argThat(isStatusRunning<Foo>(count: 0))),
        fooEffectOff(),
        enterState(argThat(isStateOf<Bar>(count: 1)), argThat(isSession())),
        watcher(argThat(isStatusRunning<Bar>(count: 1))),
        barEffectOff(),
        enterState(argThat(isStateOf<Baz>(count: 2)), argThat(isSession())),
        watcher(argThat(isStatusRunning<Baz>(count: 2))),
        bazEffectOff(),
        enterState(argThat(isStateOf<Foo>(count: 3)), argThat(isSession())),
      ]);
    });

    test('will execute onLeave function on machine dispose', () {
      final fooOff = MockEffectDispose();

      final m = TestMachine(
        configuredEnterState: (machine, state, session) =>
            state.mapOrNull<OffState?>(
          foo: (value) {
            session.onLeave(fooOff);
          },
        ),
      );
      m.dispose();
      verify(fooOff()).called(1);
    });

    test(
        'will execute onLeave function on machine dispose even if machine is disposed in effect',
        () {
      final off = MockEffectDispose();
      late TestMachine m;
      m = TestMachine(
        configuredEnterState: (machine, state, session) => state.mapOrNull(
          bar: (value) {
            session.onLeave(off);
            m.dispose();
          },
        ),
      );
      assumeInState<Foo>(m).next(1);
      verify(off()).called(1);
    });

    test('allows to sync execute an event', () {
      final m = TestMachine(
        configuredEnterState: (machine, state, session) =>
            state.mapOrNull<OffState?>(
          foo: (state) {
            state.next(1);
          },
          bar: (state) {
            state.next(1);
          },
        ),
      );

      final watcher = MockWatch();
      final disp = m.watch(watcher, fireImmediately: false);

      expect(m.status, isStatusRunning<Baz>(count: 2));

      assumeInState<Baz>(m).next(1);

      expect(m.status, isStatusRunning<Baz>(count: 5));

      verify(watcher(argThat(isStatusRunning<Baz>(count: 5)))).called(1);
      verifyNoMoreInteractions(watcher);
      disp();
    });

    group('listen to event:', () {
      TypeMatcher<StateMachineEvent> isEventOf(
        String name, {
        Object? payload,
      }) {
        var matcher = isA<StateMachineEvent>()
            .having((s) => s.name, 'event.name', equals(name));
        if (null != payload) {
          matcher =
              matcher.having((s) => s.data, 'event.data', equals(payload));
        }

        return matcher;
      }

      test('listener callback receives corrent values', () {
        final listener = MockEventListenerInSession();
        final m = TestMachine(
          configuredEnterState: (machine, state, session) {
            session.listenToEvent((event, willTransition) {
              listener(event, willTransition, state, session);
            });
          },
        );

        assumeInState<Foo>(m).next(1);
        m.send(StateMachineEvent('other', 'cool'));
        assumeInState<Bar>(m).next(2);

        verifyInOrder([
          listener(argThat(isEventOf('next', payload: 1)), argThat(isTrue),
              argThat(isStateOf<Foo>()), argThat(isSession())),
          listener(
              argThat(isEventOf('other', payload: 'cool')),
              argThat(isFalse),
              argThat(isStateOf<Bar>()),
              argThat(isSession())),
          listener(argThat(isEventOf('next', payload: 2)), argThat(isTrue),
              argThat(isStateOf<Bar>()), argThat(isSession())),
        ]);
        verifyNoMoreInteractions(listener);
      });

      test('sync send event during callback', () {
        final watcher = MockWatch();
        final m = TestMachine(
          configuredEnterState: (machine, state, session) {
            session.listenToEvent((event, willTransition) {
              if (state is Foo && event.name == 'justToTriggerTheEvent') {
                state.next(1);
              }
            });
          },
        );
        final d = m.watch(watcher, fireImmediately: false);
        m.send(StateMachineEvent('justToTriggerTheEvent', null));

        verify(watcher(argThat(isStatusRunning<Bar>(count: 1))));
        verifyNoMoreInteractions(watcher);

        d();
      });
    });
  });

  group('watch', () {
    test('hasListener', () {
      final m = TestMachine();

      expect(m.hasListener, isFalse);
      final d1 = m.watch((value) {});
      final d2 = m.watch((value) {});
      expect(m.hasListener, isTrue);
      d2();
      expect(m.hasListener, isTrue);
      d1();
      expect(m.hasListener, isFalse);
      m.watch((value) {});
      expect(m.hasListener, isTrue);
      m.dispose();
      expect(() => m.hasListener, isDisposedError);
    });

    test('immediately receive current value on watch', () {
      final m = TestMachine();
      final watcher = MockWatch();
      m.watch(watcher);

      verify(watcher(argThat(isStatusRunning<Foo>()))).called(1);
    });

    test('skip receiving value on watch', () {
      final m = TestMachine();
      final watcher = MockWatch();
      m.watch(watcher, fireImmediately: false);
      verifyNever(watcher(argThat(isA<TestState>())));
    });

    test('receives updates', () {
      final m = TestMachine();
      final watcher = MockWatch();
      m.watch(watcher);

      assumeInState<Foo>(m).next(1);
      assumeInState<Bar>(m).next(1);
      verifyInOrder([
        watcher(argThat(isStatusRunning<Foo>(count: 0))),
        watcher(argThat(isStatusRunning<Bar>(count: 1))),
        watcher(argThat(isStatusRunning<Baz>(count: 2)))
      ]);
    });

    test('dispose machine and remove listener after will not throw', () {
      final m = TestMachine();
      final watcher = MockWatch();
      final disp = m.watch(watcher);
      m.dispose();
      expect(() => disp(), isNot(throwsException));
    });

    test('cannot subscribe on disposed machine', () {
      expect(
          () => TestMachine()
            ..dispose()
            ..watch((_) {}),
          isDisposedError);
    });
  });
}

class TestMachine extends StateMachine<TestState> {
  TestState createFoo({required SendEvent send, required int count}) {
    return TestState.foo(
      count,
      (int incr) {
        send(StateMachineEvent('next', incr));
      },
    );
  }

  TestState createBar({required SendEvent send, required int count}) {
    return TestState.bar(
      count,
      (int incr) {
        send(StateMachineEvent('next', incr));
      },
    );
  }

  TestState createBaz({required SendEvent send, required int count}) {
    return TestState.baz(
      count,
      (int incr) {
        send(StateMachineEvent('next', incr));
      },
    );
  }

  static TestState defaultTransition(TestMachine machine, TestState state,
      StateMachineEvent event, SendEvent send) {
    if (event.name != 'next') return state;
    return state.map(
      foo: (foo) => machine.createBar(
        send: send,
        count: (event.data as int) + foo.count,
      ),
      bar: (bar) => machine.createBaz(
        send: send,
        count: (event.data as int) + bar.count,
      ),
      baz: (baz) => machine.createFoo(
        send: send,
        count: (event.data as int) + baz.count,
      ),
    );
  }

  TestMachine({
    this.configuredEnterState,
    this.configuredTransition,
    this.configuredInitialState,
    bool autostart = true,
  }) : super(autostart: autostart);

  final TestState Function(TestMachine machine, TestState state,
      StateMachineEvent event, SendEvent send)? configuredTransition;

  final void Function(
          TestMachine machine, TestState state, StateMachineSession session)?
      configuredEnterState;

  final TestState Function(TestMachine machine, SendEvent send)?
      configuredInitialState;

  StateMachineSession get debugSession => session;

  @override
  TestState getInitialState(SendEvent send) {
    return configuredInitialState?.call(this, send) ??
        createFoo(send: send, count: 0);
  }

  @override
  TestState transition(
          TestState state, StateMachineEvent event, SendEvent send) =>
      configuredTransition?.call(this, state, event, send) ??
      defaultTransition(this, state, event, send);

  @override
  void enterState(TestState state, StateMachineSession session) {
    configuredEnterState?.call(this, state, session);
  }

  void send(StateMachineEvent event) {
    session.send(event);
  }
}
